import 'package:flutter/material.dart';
import 'package:rxdart/rxdart.dart';

import 'package:tbloc_dart/core/base/base.dart';
import 'package:tbloc_dart/core/events/events.dart';
import 'package:tbloc_dart/core/states/states.dart';
import 'package:tbloc_dart/core/types/types.dart';

abstract class BidirectionalBloc<E extends BlocEvent, S extends BlocState>
    extends Bloc<S> {
  @protected
  final PublishSubject<BlocEvent> internalEventController =
      PublishSubject<BlocEvent>();
  @protected
  final PublishSubject<Object> errorController = PublishSubject<Object>();
  @protected
  final PublishSubject<E> externalEventController = PublishSubject<E>();
  @protected
  Stream<S> mapEventToState(E event, S currentState);

  Stream<E> get onEvent => externalEventController.stream;

  Stream<Object> get onError => errorController.stream;

  Function(BlocEvent) get dispatchEvent {
    if (!internalEventController.isClosed) {
      return internalEventController.sink.add;
    }

    return _dispatchEvent;
  }

  BidirectionalBloc({
    S initialState,
    BlocStateBuilder<S> stateBuilder,
  }) : super(
          initialState: initialState,
          stateBuilder: stateBuilder,
        ) {
    internalEventController
        .asyncExpand((BlocEvent event) {
          if (event.error != null) {
            throw (event.error);
          } else if (event.resetState) {
            return Stream.value(getInitialState());
          }

          if (event is E) {
            externalEventController.sink.add(event);

            final currentState = stateController.value ?? initialState;
            return mapEventToState(event, currentState);
          }

          return Stream.value(currentState);
        })
        .handleError(handleError)
        .listen((S nextState) {
          setState(nextState);
        });
  }

  void reset() => dispatchEvent(BlocEvent(resetState: true));

  @override
  void dispose() {
    internalEventController.close();
    externalEventController.close();
    errorController.close();
    super.dispose();
  }

  @protected
  void handleError(Object error) => errorController.sink.add(error);

  void _dispatchEvent(BlocEvent event) {}
}
